/*
 * Copyright (C) 2005 by the gtk2-perl team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: IdlFunctionInfo.xs,v 1.1.1.1 2005/08/21 14:39:45 torsten Exp $
 */

#include "gobject-introspection-perl.h"

extern GType gidlperl_get_gtype (GIRegisteredTypeInfo *info);

static void
sv_to_arg (SV * sv,
	   GArgument * arg,
	   GITypeInfo * info,
           GIArgInfo * arg_info)
{
	GITypeTag tag = g_type_info_get_tag (info);

	if (!sv || !SvOK (sv))
		if (!g_arg_info_may_be_null (arg_info))
			croak ("undefined value for a mandatory argument encountered");

	switch (tag) {
	    case GI_TYPE_TAG_VOID:
		/* do nothing */
		break;

	    case GI_TYPE_TAG_BOOLEAN:
		arg->v_boolean = SvTRUE (sv);
		break;

	    case GI_TYPE_TAG_INT8:
		arg->v_int8 = (char) SvIV (sv);
		break;

	    case GI_TYPE_TAG_UINT8:
		arg->v_uint8 = (unsigned char) SvUV (sv);
		break;

	    case GI_TYPE_TAG_INT16:
		arg->v_int16 = (gint16) SvIV (sv);
		break;

	    case GI_TYPE_TAG_UINT16:
		arg->v_uint16 = (guint16) SvUV (sv);
		break;

	    case GI_TYPE_TAG_INT32:
		arg->v_int32 = (gint32) SvIV (sv);
		break;

	    case GI_TYPE_TAG_UINT32:
		arg->v_uint32 = (guint32) SvUV (sv);
		break;

	    case GI_TYPE_TAG_INT64:
		croak ("FIXME - GI_TYPE_TAG_INT64");
		break;

	    case GI_TYPE_TAG_UINT64:
		croak ("FIXME - GI_TYPE_TAG_UINT64");
		break;

	    case GI_TYPE_TAG_FLOAT:
		arg->v_float = (float) SvNV (sv);
		break;

	    case GI_TYPE_TAG_DOUBLE:
		arg->v_double = SvNV (sv);
		break;

	    case GI_TYPE_TAG_INT:
		arg->v_int = SvIV (sv);
		break;

	    case GI_TYPE_TAG_UINT:
		arg->v_uint = SvUV (sv);
		break;

	    case GI_TYPE_TAG_LONG:
		arg->v_long = SvIV (sv);
		break;

	    case GI_TYPE_TAG_ULONG:
		arg->v_ulong = SvUV (sv);
		break;

	    case GI_TYPE_TAG_ARRAY:
		croak ("FIXME - GI_TYPE_TAG_ARRAY");
		break;

	    case GI_TYPE_TAG_INTERFACE:
	    {
		GIBaseInfo *interface = g_type_info_get_interface (info);
		GIInfoType info_type = g_base_info_get_type (interface);

		switch (info_type) {
		    case GI_INFO_TYPE_OBJECT:
			arg->v_pointer = SvGObject (sv);
			break;

		    case GI_INFO_TYPE_BOXED:
		    {
			GType type = gidlperl_get_gtype ((GIRegisteredTypeInfo *) interface);
			arg->v_pointer = gperl_get_boxed_check (sv, type);
		    }
			break;

		    case GI_INFO_TYPE_ENUM:
		    {
			GType type = gidlperl_get_gtype ((GIRegisteredTypeInfo *) interface);
			arg->v_pointer = GINT_TO_POINTER (gperl_convert_enum (type, sv));
		    }
			break;

		    case GI_INFO_TYPE_UNRESOLVED:
		    {
			const gchar *type_name = g_base_info_get_name (interface);
			GType type = g_type_from_name (type_name);

			if (!type)
				croak ("Could not determine type for %s", type_name);

			if (g_type_is_a (type, G_TYPE_OBJECT)) {
				arg->v_pointer = SvGObject (sv);
			} else if (g_type_is_a (type, G_TYPE_BOXED)) {
				arg->v_pointer = gperl_get_boxed_check (sv, type);
			} else {
				croak ("Don't know how to handle type %d (%s)",
				       type, type_name);
			}
		    }

			break;

		    default:
			croak ("Don't know how to handle info type %d", info_type);
			break;
		}
	    }
		break;

	    case GI_TYPE_TAG_GLIST:
		croak ("FIXME - GI_TYPE_TAG_GLIST");
		break;

	    case GI_TYPE_TAG_GSLIST:
		croak ("FIXME - GI_TYPE_TAG_GSLIST");
		break;

	    case GI_TYPE_TAG_GHASH:
		croak ("FIXME - GI_TYPE_TAG_GHASH");
		break;

	    case GI_TYPE_TAG_ERROR:
		croak ("FIXME - A GError as an in/inout arg?  Should never happen!");
		break;

	    case GI_TYPE_TAG_SSIZE:
		croak ("FIXME - GI_TYPE_TAG_SSIZE");
		break;

	    case GI_TYPE_TAG_SIZE:
		croak ("FIXME - GI_TYPE_TAG_SIZE");
		break;

	    case GI_TYPE_TAG_UTF8:
		arg->v_string = SvOK (sv) ? SvGChar (sv) : NULL;
		break;

	    case GI_TYPE_TAG_FILENAME:
		arg->v_string = SvOK (sv) ? gperl_filename_from_sv (sv) : NULL;
		break;

	    default:
		croak ("Unhandled info tag %d", tag);
	}
}

static SV *
arg_to_sv (const GArgument * arg,
	   GITypeInfo * info,
           GITransfer transfer)
{
	GITypeTag tag = g_type_info_get_tag (info);
	gboolean own = transfer == GI_TRANSFER_EVERYTHING;

	switch (tag) {
	    case GI_TYPE_TAG_VOID:
		return NULL;

	    case GI_TYPE_TAG_BOOLEAN:
		return boolSV (arg->v_boolean);

	    case GI_TYPE_TAG_INT8:
		return newSViv (arg->v_int8);

	    case GI_TYPE_TAG_UINT8:
		return newSVuv (arg->v_uint8);
		break;

	    case GI_TYPE_TAG_INT16:
		return newSViv (arg->v_int16);
		break;

	    case GI_TYPE_TAG_UINT16:
		return newSVuv (arg->v_uint16);
		break;

	    case GI_TYPE_TAG_INT32:
		return newSViv (arg->v_int32);
		break;

	    case GI_TYPE_TAG_UINT32:
		return newSVuv (arg->v_uint32);
		break;

	    case GI_TYPE_TAG_INT64:
		croak ("FIXME - GI_TYPE_TAG_INT64");
		break;
	    case GI_TYPE_TAG_UINT64:
		croak ("FIXME - GI_TYPE_TAG_UINT64");
		break;

	    case GI_TYPE_TAG_FLOAT:
		return newSVnv (arg->v_float);
		break;

	    case GI_TYPE_TAG_DOUBLE:
		return newSVnv (arg->v_double);
		break;

	    case GI_TYPE_TAG_INT:
		return newSViv (arg->v_int);
		break;

	    case GI_TYPE_TAG_UINT:
		return newSVuv (arg->v_uint);
		break;

	    case GI_TYPE_TAG_LONG:
		return newSViv (arg->v_long);
		break;

	    case GI_TYPE_TAG_ULONG:
		return newSVuv (arg->v_ulong);
		break;

	    case GI_TYPE_TAG_ARRAY:
		croak ("FIXME - GI_TYPE_TAG_ARRAY");
		break;

	    case GI_TYPE_TAG_INTERFACE:
	    {
		GIBaseInfo *interface = g_type_info_get_interface (info);
		GIInfoType info_type = g_base_info_get_type (interface);

		switch (info_type) {
		    case GI_INFO_TYPE_OBJECT:
			return gperl_new_object (arg->v_pointer, own);
			break;

		    case GI_INFO_TYPE_BOXED:
		    {
			GType type = gidlperl_get_gtype ((GIRegisteredTypeInfo *) interface);
			return gperl_new_boxed (arg->v_pointer, type, own);
		    }
			break;

		    case GI_INFO_TYPE_ENUM:
		    {
			GType type = gidlperl_get_gtype ((GIRegisteredTypeInfo *) interface);
			return gperl_convert_back_enum (type, GPOINTER_TO_INT (arg->v_pointer));
		    }
			break;

		    case GI_INFO_TYPE_UNRESOLVED:
		    {
			const gchar *type_name = g_base_info_get_name (interface);
			GType type = g_type_from_name (type_name);

			if (!type)
				croak ("Could not determine type for %s", type_name);

			if (g_type_is_a (type, G_TYPE_OBJECT)) {
				GObject *object = arg->v_pointer;
				return gperl_new_object (object, own);
			} else if (g_type_is_a (type, G_TYPE_BOXED)) {
				return gperl_new_boxed (arg->v_pointer, type, own);
			} else {
				croak ("Don't know how to handle type %d (%s)",
				       type, type_name);
			}
		    }

			break;

		    default:
			croak ("Don't know how to handle info type %d", info_type);
			break;
		}
	    }

		break;

	    case GI_TYPE_TAG_GLIST:
	    {
		GITypeInfo *param_info;
		GList *i;
		AV *av;
		SV *value;

		param_info = g_type_info_get_param_type (info, 0);
		av = newAV ();

		for (i = arg->v_pointer; i; i = i->next) {
			value = arg_to_sv (arg, param_info, transfer);
			if (value)
				av_push (av, value);
		}

		return newRV_noinc ((SV *) av);
	    }
		break;

	    case GI_TYPE_TAG_GSLIST:
		croak ("FIXME - GI_TYPE_TAG_GSLIST");
		break;

	    case GI_TYPE_TAG_GHASH:
		croak ("FIXME - GI_TYPE_TAG_GHASH");
		break;

	    case GI_TYPE_TAG_ERROR:
		if (arg->v_pointer)
			gperl_croak_gerror (NULL, arg->v_pointer);
		break;

	    case GI_TYPE_TAG_SSIZE:
		croak ("FIXME - GI_TYPE_TAG_SSIZE");
		break;

	    case GI_TYPE_TAG_SIZE:
		croak ("FIXME - GI_TYPE_TAG_SIZE");
		break;

	    case GI_TYPE_TAG_UTF8:
		return newSVGChar (arg->v_string);
		break;

	    case GI_TYPE_TAG_FILENAME:
		return gperl_sv_from_filename (arg->v_string);
		break;

	    default:
		croak ("Unhandled info tag %d", tag);
	}

	return NULL;
}

/* ------------------------------------------------------------------------- */

SV *
newSVGIFunctionInfo (GIFunctionInfo *info)
{
        SV *sv = newSV (0);

        return sv_setref_pv (sv, "Glib::Object::Introspection::FunctionInfo", info);
}

GIFunctionInfo *
SvGIFunctionInfo (SV *info)
{
        return INT2PTR (GIFunctionInfo *, SvIV (SvRV (info)));

}

/* ------------------------------------------------------------------------- */

XS(XS_Glib__Idl__FunctionInfo_invoke)
{
	dXSARGS;

	if (items < 1)
		Perl_croak(aTHX_ "Usage: Glib::Object::Introspection::FunctionInfo::invoke(info, ...)");

	SP -= items;

    {
	GIFunctionInfo * info = SvGIFunctionInfo (ST(0));
	int have_args;
	int n_args;
	int n_in_args;
	int n_out_args;
	int i, out_i;
	GITypeInfo ** out_arg_type = NULL;
	GITypeInfo * return_type_info = NULL;
	GITransfer return_type_transfer;
	gboolean has_return_value;
	GArgument return_value;
	GArgument * in_args = NULL;
	GArgument * out_args = NULL;
	GError * error = NULL;

#define OFFSET 1
	have_args = items - OFFSET;
	n_args = g_callable_info_get_n_args ((GICallableInfo *) info);

	/* to allow us to make only one pass through the arg list, allocate
	 * enough space for all args in both the out and in lists.  we'll
	 * only use as much as we need.  since function argument lists are
	 * typically small, this shouldn't be a big problem. */
	if (n_args) {
		in_args = gperl_alloc_temp (sizeof (GArgument) * n_args);
		out_args = gperl_alloc_temp (sizeof (GArgument) * n_args);
		out_arg_type = gperl_alloc_temp (sizeof (GITypeInfo*) * n_args);
	}

	n_in_args = n_out_args = 0;
	for (i = 0 ; i < n_args ; i++) {
		GIArgInfo * arg_info =
			g_callable_info_get_arg ((GICallableInfo *) info, i);
		GITypeInfo * arg_type = g_arg_info_get_type (arg_info);

		switch (g_arg_info_get_direction (arg_info)) {
		    case GI_DIRECTION_IN:
			sv_to_arg (ST (i+OFFSET), in_args+n_in_args++, arg_type, arg_info);
			g_base_info_unref ((GIBaseInfo *) arg_type);
			break;
		    case GI_DIRECTION_OUT:
			out_args[n_out_args++].v_pointer =
				gperl_alloc_temp (sizeof (GArgument));
			out_arg_type[n_out_args-1] = arg_type;
			break;
		    case GI_DIRECTION_INOUT:
		    {
			GArgument * temp =
				gperl_alloc_temp (sizeof (GArgument));
			sv_to_arg (ST (i+OFFSET), temp, arg_type, arg_info);
			in_args[n_in_args++].v_pointer =
				out_args[n_out_args++].v_pointer =
					temp;
			out_arg_type[n_out_args-1] = arg_type;
		    }
			break;
		}
		g_base_info_unref ((GIBaseInfo*) arg_info);
	}
#undef OFFSET

	/* if (n_in_args != have_args)
		croak ("%s needs %d args, got %d",
		       g_function_info_get_symbol (info),
		       n_in_args, have_args); */

	if (! g_function_info_invoke (info, in_args, n_in_args,
				      out_args, n_out_args,
				      &return_value, &error))
		gperl_croak_gerror (NULL, error);

	/*
	 * place return value and output args on the stack
	 */
	return_type_info = g_callable_info_get_return_type ((GICallableInfo *) info);
	return_type_transfer = g_callable_info_get_caller_owns ((GICallableInfo *) info);
	has_return_value = GI_TYPE_TAG_VOID != g_type_info_get_tag (return_type_info);

	if (has_return_value) {
		SV *value = arg_to_sv (&return_value,
		                       return_type_info,
		                       return_type_transfer);
		if (value)
			XPUSHs (sv_2mortal (value));
	}

	/* out args */
	for (i = out_i = 0 ; i < n_args ; i++) {
		GIArgInfo * arg_info =
			g_callable_info_get_arg ((GICallableInfo *) info, i);

		switch (g_arg_info_get_direction (arg_info)) {
		    case GI_DIRECTION_OUT:
		    case GI_DIRECTION_INOUT:
		    {
			GITransfer transfer = g_arg_info_get_ownership_transfer (arg_info);
			SV *sv = arg_to_sv (out_args[out_i].v_pointer, out_arg_type[out_i], transfer);
			if (sv)
				XPUSHs (sv_2mortal (sv));
			g_base_info_unref ((GIBaseInfo*) out_arg_type[out_i]);
		    }
			out_i++;
			break;
		    default:
			break;
		}
	}

	if (has_return_value)
		out_i++;

	if (out_i == 0) {
		XSRETURN_EMPTY;
	} else if (out_i == 1) {
		XSRETURN (1);
	} else {
		PUTBACK;
		return;
	}
    }
}

MODULE = Glib::Object::Introspection::FunctionInfo	PACKAGE = Glib::Object::Introspection::FunctionInfo	PREFIX = g_function_info_

BOOT:
	newXS("Glib::Object::Introspection::FunctionInfo::invoke", XS_Glib__Idl__FunctionInfo_invoke, file);

const gchar * g_function_info_get_symbol (GIFunctionInfo *info);

GIFunctionInfoFlags g_function_info_get_flags (GIFunctionInfo *info);

GIPropertyInfo * g_function_info_get_property (GIFunctionInfo *info);

GIVFuncInfo * g_function_info_get_vfunc (GIFunctionInfo *info);
